/*-----------------------------------------------------------------------

File  : ccl_clauses.c

Author: Stephan Schulz

Contents
 
  Clauses - Infrastructure functions

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Thu Apr 16 19:38:16 MET DST 1998
    New

<2> Wed Oct 14 08:40:21 MET DST 1998
    Extended (ClauseNotGreaterEqual, foundEqLitLater),
         Joachim Steinbach

-----------------------------------------------------------------------*/


#include "ccl_clauses.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

static long global_clause_counter = 0; /* Counts calls to
					  ClauseAlloc()  and
					  ClauseCopy */

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: foundEqLitLater
//
// Checks whether there is a literal (say L) in lits2 which is
// equivalent wrt. an ordering (given in ocb) to the first literal in
// lits1 (say M). In this case, the property `EPHasEquiv' is set for L
// and M.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool foundEqLitLater(OCB_p ocb, Eqn_p lits1, Eqn_p lits2)
{
   while (lits2)
   {
      if ((!EqnHasEquiv(lits2)) &&
	  (LiteralCompare(ocb, lits1, lits2) == to_equal))
      {
	 EqnSetProp(lits1, EPHasEquiv);
	 EqnSetProp(lits2, EPHasEquiv);
	 return true;
      }
      lits2 = lits2->next;
   }
   return false;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: EmptyClauseAlloc()
//
//   Return a pointer to an empty clause initialized with rational
//   values.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

Clause_p EmptyClauseAlloc()
{
   Clause_p handle = ClauseCellAlloc();

   handle->literals    = NULL;
   handle->neg_lit_no  = 0;
   handle->pos_lit_no  = 0;
   handle->evaluations = NULL;
   handle->set         = NULL;
   handle->properties  = CPIgnoreProps;
   handle->proof_depth = 0;
   handle->proof_size  = 0;
   handle->children    = NULL;
   handle->parent1     = NULL;
   handle->parent2     = NULL;
   handle->pred        = NULL;
   handle->succ        = NULL;

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseAlloc()
//
//   Create a new clause with the literal list list.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Clause_p ClauseAlloc(Eqn_p literals)
{
   Clause_p handle = EmptyClauseAlloc();

   handle->literals = literals;
   handle->ident = ++global_clause_counter;

   while(literals)
   {
      if(EqnIsPositive(literals))
      {
	 handle->pos_lit_no++;
      }
      else
      {
	 handle->neg_lit_no++;
      }
      literals = literals->next;
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseFree()
//
//   Free a clause. Does not take care of parents, children, etc., but
//   releases the memory directly taken by the clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseFree(Clause_p junk)
{
   assert(!junk->set);
   EvalListFree(junk->evaluations);   
   EqnListFree(junk->literals);
   PTreeFree(junk->children);
   ClauseCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseCopy()
//
//   Create a semantically equivalent clause and return a pointer to
//   it. Does not copy parents, children, etc. Terms in the original
//   clause are interpreted as instantiated, and are created in or
//   retrived from the new bank. Evaluations are not copied. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Clause_p ClauseCopy(Clause_p clause, TB_p bank)
{
   Clause_p handle = ClauseCellAlloc();

   handle->literals    = EqnListCopy(clause->literals, bank);
   handle->ident       = ++global_clause_counter;
   handle->neg_lit_no  = clause->neg_lit_no;
   handle->pos_lit_no  = clause->pos_lit_no;
   handle->evaluations = NULL;
   handle->set         = NULL;
   handle->properties  = clause->properties;
   handle->proof_depth = clause->proof_depth;
   handle->proof_size  = clause->proof_size;
   handle->children    = NULL;
   handle->parent1     = NULL;
   handle->parent2     = NULL;
   handle->pred        = NULL;
   handle->succ        = NULL;

   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: ClausePrintList()
//
//   Print a clause as a declarative list of literals.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintList(FILE* out, Clause_p clause, bool fullterms)
{
   EqnListPrint(out, clause->literals, "; ", normal, fullterms);
   fprintf(out, " <-.");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintAxiom()
//
//   Print a clause as a declarative axiom in normal form, i.e. print
//   positive literals first, then <-, then negative literals.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintAxiom(FILE* out, Clause_p clause, bool fullterms)
{
   Eqn_p handle;
   int   i;

   i = 0;
   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(EqnIsPositive(handle))
      {
	 EqnPrint(out, handle, normal, fullterms);
	 i++;
	 if(i<clause->pos_lit_no)
	 {
	    fprintf(out, "; ");
	 }
      }
   }
   fprintf(out, " <- ");
   
   i = 0;
   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(!EqnIsPositive(handle))
      {
	 EqnPrint(out, handle, special, fullterms);
	 i++;
	 if(i<clause->neg_lit_no)
	 {
	    fprintf(out, ", ");
	 }
      }
   }
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintRule()
//
//   Print a clause as a rule, with the head literal as the conclusion
//   and the remaining literals as preconditions.  If a
//   clause has a single literal only, print it as a fact.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintRule(FILE* out, Clause_p clause, bool fullterms)
{
   if(clause->literals)
   {
      EqnPrint(out, clause->literals, normal, fullterms);
      if(clause->literals->next)
      {
	 fprintf(out, " <- ");
	 EqnListPrint(out, clause->literals->next, ", ", special,
		      fullterms);
      }
   }
   else
   {
      fprintf(out, " <- ");
   }
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintGoal()
//
//   Print a clause as a goal, i.e. put all literals behind <-.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintGoal(FILE* out, Clause_p clause, bool fullterms)
{
   fprintf(out, "<- ");
   EqnListPrint(out, clause->literals, ", ", special, fullterms);
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintQuery()
//
//   Print a clause as a procedural query, i.e. put all literals
//   behind ?-.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintQuery(FILE* out, Clause_p clause, bool fullterms)
{
   fprintf(out, "?- ");
   EqnListPrint(out, clause->literals, ", ", special, fullterms);
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrint()
//
//   Print a clause in the most canonical representation.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrint(FILE* out, Clause_p clause, bool fullterms)
{
   DEBUG(65535, fprintf(out, "/*(%3ld)*/ ", clause->ident););

   /* fprintf(out, "(%ld, %ld)", clause->proof_depth,
      clause->proof_size); */
   if(clause->pos_lit_no)
   {
      ClausePrintAxiom(out, clause, fullterms);
   }
   else
   {
      ClausePrintQuery(out, clause, fullterms);
   }
   DEBUG(64, EvalListPrintComment(out, clause->evaluations););
   DEBUG(64, printf("Address: %d", (int)clause););
}


/*-----------------------------------------------------------------------
//
// Function: ClauseStartsMaybe()
//
//   Return true if a clause possibly starts on the current position
//   in the input, i.e. if TermStartToken, TildeSign, ?-, or <- are
//   present on the input stream.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseStartsMaybe(Scanner_p in)
{
   if(TestInpTok(in, TermStartToken|TildeSign))
   {
      return true;
   }
   if(TestInpTok(in, LesserSign|QuestionMark))
   {
      if(!LookToken(in,1)->skipped && TestTok(LookToken(in, 1), Hyphen))
      {
	 return true;
      }
   }
   return false;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseParse()
//
//   Parse a clause.
//
// Global Variables: -
//
// Side Effects    : Input, memory operations
//
/----------------------------------------------------------------------*/

Clause_p ClauseParse(Scanner_p in, TB_p bank)
{
   Eqn_p    concl, precond;
   bool     procedural = false;
   Clause_p handle;

   VarBankClearExtNames(bank->vars);
   
   concl = EqnListParse(in, bank, Semicolon);
   if(TestInpTok(in, Colon))
   {
      if(EqnListLength(concl)>1)
      {
	 AktTokenError(in, 
		       "Procedural rule cannot have more than one "
		       "head literal",
		       false);
      }
      procedural = true;
   }
   else if(TestInpTok(in, QuestionMark))
   {
      if(EqnListLength(concl)>0)
      {
	 AktTokenError(in, 
		       "Query should consist only of tail literals",
		       false);
      }
      procedural = true;
   }
   if(TestInpTok(in, Fullstop))
   {
      if(EqnListLength(concl)>1)
      {
	 AktTokenError(in, 
		       "Procedural fact cannot have more than one "
		       "literal",
		       false);
      }
      procedural = true;
   }
   else
   {
      AcceptInpTok(in, LesserSign | Colon | QuestionMark);
      AcceptInpTokNoSkip(in, Hyphen);
      
      precond = EqnListParse(in, bank, Comma);
   
      if(procedural && EqnListLength(precond)==0)
      {
	  AktTokenError(in, 
			"Procedural rule or query needs at least one "
			"tail literal (Hey! I did not make this "
			" syntax! -StS)", 
			false);
      }
      EqnListNegateEqns(precond);
      EqnListAppend(&concl, precond);
   }
   AcceptInpTok(in, Fullstop);
   handle = ClauseAlloc(concl);
   if(ClauseIsGoal(handle))
   {
      ClauseSetProp(handle, CPInSOS);
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseMarkMaximalTerms()
//
//   Orient literals, mark maximal literals.
//
// Global Variables: -
//
// Side Effects    : Changes clause structure, orients equations
//
/----------------------------------------------------------------------*/

void ClauseMarkMaximalTerms(OCB_p ocb, Clause_p clause)
{
   EqnListOrient(ocb, clause->literals);
   EqnListMaximalLiterals(ocb, clause->literals);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseMarkRestrictedTerms()
//
//   Mark terms on which rewriting is restricted.
//
// Global Variables: -
//
// Side Effects    : Changes term bank...
//
/----------------------------------------------------------------------*/

void ClauseMarkRestrictedTerms(Clause_p clause)
{
   EqnListMarkRestrictedTerms(clause->literals);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseDetachParents()
//
//   Discouple a clause from it's parents, i.e. make the parents
//   forget the clause and vice versa.
//
// Global Variables: -
//
// Side Effects    : Changes parens children-tree.
//
/----------------------------------------------------------------------*/

void ClauseDetachParents(Clause_p clause)
{
   bool result;

   if(clause->parent1)
   {
      result = PTreeDeleteEntry(&(clause->parent1->children), clause);
      assert(result);
      clause->parent1 = NULL;
   }
   if(clause->parent2)
   {
      result = PTreeDeleteEntry(&(clause->parent2->children), clause);
      assert(result);
      clause->parent2 = NULL;
   }   
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRegisterChild()
//
//   Make child a child of clause. Clause already should be a parent
//   to child!
//
// Global Variables: -
//
// Side Effects    : Changes clause->children
//
/----------------------------------------------------------------------*/

void ClauseRegisterChild(Clause_p clause, Clause_p child)
{
   assert(child->parent1==clause || child->parent2==clause);
   
   PTreeStore(&(clause->children), child);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveSuperfluousLiterals()
//
//   Remove duplicate and trivial negative literals from the
//   clause. Return number of removed literals.
//
// Global Variables: -
//
// Side Effects    : Changes clause, termbanks
//
/----------------------------------------------------------------------*/

int ClauseRemoveSuperfluousLiterals(Clause_p clause)
{
   Eqn_p handle;
   int   removed = 0;

   removed += EqnListRemoveResolved(&(clause->literals), TBTermEqual);
   removed += EqnListRemoveDuplicates(clause->literals, TBTermEqual);
   
   clause->neg_lit_no = 0;
   clause->pos_lit_no = 0;   
   handle = clause->literals;

   while(handle)
   {
      if(EqnIsPositive(handle))
      {
	 clause->pos_lit_no++;
      }
      else
      {
	 clause->neg_lit_no++;
      }
      handle = handle->next;
   }
   return removed;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseAddEvalCell()
//
//   Add an evaluation cell (as the first evaluation) to a clause. 
//
// Global Variables: -
//
// Side Effects    : Changes eval list, makes evaluation->object point
//                   to clause.
//
/----------------------------------------------------------------------*/

void ClauseAddEvalCell(Clause_p clause, Eval_p evaluation)
{
   evaluation->object  = clause;
   evaluation->next    = clause->evaluations;
   clause->evaluations = evaluation;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveEvaluations()
//
//   Remove the evaluations from a clause and free the EvalCells.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseRemoveEvaluations(Clause_p clause)
{
   EvalListFree(clause->evaluations);
   clause->evaluations = NULL;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseWeight()
//
//   Compute the weight of a clause by counting function symbols and
//   variables and applying various modifiers.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double ClauseWeight(Clause_p clause, double max_term_multiplier,
		    double max_literal_multiplier, double
		    pos_multiplier, long vweight, long fweight, bool
		    count_eq_encoding)
{
   Eqn_p  handle;
   double res = 0;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      res += LiteralWeight(handle, max_term_multiplier,
			   max_literal_multiplier, pos_multiplier,
			   vweight, fweight, count_eq_encoding);
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseOrientWeight()
//
//   Compute the weight of a clause by counting function symbols and
//   variables and applying various other modifiers ;-).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double ClauseOrientWeight(Clause_p clause, double
			  unorientable_literal_multiplier,
			  double max_literal_multiplier, double
			  pos_multiplier, long vweight, long fweight, bool
			  count_eq_encoding)
{
   Eqn_p  handle;
   double res = 0 ,tmp;
   
   for(handle = clause->literals; handle; handle = handle->next)
   {
      tmp = LiteralWeight(handle, 1, max_literal_multiplier,
			  pos_multiplier, vweight, fweight,
			  count_eq_encoding);
      if(!EqnIsOriented(handle))
      {
	 tmp = tmp*unorientable_literal_multiplier;
      }
      res += tmp;
   }
   return res;


}


/*-----------------------------------------------------------------------
//
// Function: ClauseNotGreaterEqual
//
//   Checks whether the multiset of literals of a clause is neither
//   greater than (wrt. the multiset extension of a specified
//   ordering) nor equal to the multiset of literals of another
//   clause.
//
//   not(A >=>= B)  iff  - not(A = B) and
//                       - for each b in B\A there does not exist an a
//                         in A\B such that a > b
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseNotGreaterEqual(OCB_p ocb,
			   Clause_p clause1, Clause_p clause2)
{
   Eqn_p cl1lits = clause1->literals;
   Eqn_p cl2lits = clause2->literals, cl2litseqs;
   bool foundeq, foundeqlater, foundgtr;

   EqnListDelProp(cl1lits, EPHasEquiv | EPDominates);

   while (cl2lits)
   {
      EqnDelProp(cl2lits, EPHasEquiv | EPIsDominated);
      foundeq = foundgtr = false;

      while (cl1lits && !EqnHasEquiv(cl2lits))
      {
	 if (!EqnHasEquiv(cl1lits))
	 {
	    switch (LiteralCompare(ocb, cl1lits, cl2lits))
	    {
	    case to_greater:
		  foundeqlater = false;
		  if (!EqnDominates(cl1lits))
		  {
		     cl2litseqs = cl2lits->next;
		     foundeqlater =
			foundEqLitLater(ocb, cl1lits, cl2litseqs);
		  }
		  if (!foundeqlater)
		  {
		     EqnSetProp(cl1lits, EPDominates);
		     EqnSetProp(cl2lits, EPIsDominated);
		     foundgtr = true;		     
		  }
		  break;
	    case to_equal:
		  EqnSetProp(cl1lits, EPHasEquiv);
		  EqnSetProp(cl2lits, EPHasEquiv);
		  foundeq = true;
		  break;
	    case to_lesser:
	    case to_uncomparable:
		  break;
	    default:
		  assert(false);
		  break;
	    }
	 }
	 cl1lits = cl1lits->next;
      }

      /* early abortion if there is neither a greater nor an equal
	 element in clause1 */
      if ((!foundeq) && (!foundgtr))
      {
	 /* printf("True\n"); */
	 return true;
      }
      cl2lits = cl2lits->next;
   }
   /* printf("False\n"); */
   return false;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
