/*-----------------------------------------------------------------------

File  : ccl_eqn.h

Author: Stephan Schulz

Contents
 
  The termpair datatype: Rules, Equations, positive and negative
  literals. 

  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Fri Mar 13 17:09:13 MET 1998
    New

-----------------------------------------------------------------------*/

#ifndef CCL_EQN

#define CCL_EQN

#include <cte_match_mgu_1-1.h>
#include <cte_replace.h>
#include <cto_orderings.h>



/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/

typedef enum 
{
   EPNoProps = 0,      /* No properties set or selected */
   EPIsPositive = 1,   /* s=t (as opposed to s!=t) */
   EPIsMaximal = 2,    /* Eqn is maximal in a clause */
   EPIsStrictlyMaximal = 4,/* Eqn is strictly maximal */
   EPIsEquLiteral = 8, /* s=t, not s=$true */
   EPIsOriented = 16,  /* s=>t  or s=t ? */
   EPHasEquiv = 32,    /* Literal has been used in multiset-comparison
			  (and found an equivalent partner) */
   EPIsDominated = 64, /* Literal is dominated by another one */
   EPDominates = 64,   /* Double use of this property in potentially
                          maximal or minimal clauses */
   EPIsUsed = 128      /* For non-injective subsumption */
}EqnProperties;


/* Basic data structure for rules, equations, literals. Terms are
   always assumed to be shared and need to be manipulated while taking
   care about references! */ 

typedef struct eqncell
{
   EqnProperties  properties;/* Positive, maximal, equational */      
   long           pick_counter; /* For subsumption */
   Term_p         lterm;
   Term_p         rterm;
   TB_p           bank;      /* Terms are from this bank */
   struct eqncell *next;     /* For lists of equations */
}EqnCell, *Eqn_p, **EqnRef;


typedef enum
{
   NoSide = 0,
   LeftSide = 1,
   MaxSide = 1,
   RightSide = 2,
   MinSide = 2,
   BothSides = 3
}EqnSide;


/* What will be parsed as the equal-predicate: */

#define EQUAL_PREDICATE "equal"

/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/


extern bool EqnUseInfix; /* s = t or EQ(s = t)  ? */
extern bool EqnFullEquationalRep; /* P(x) = $true ? */

#define EqnCellAlloc()    (EqnCell*)SizeMalloc(sizeof(EqnCell))
#define EqnCellFree(junk) SizeFree(junk, sizeof(EqnCell))

Eqn_p   EqnAlloc(Term_p lterm, Term_p rterm, TB_p bank, bool positive);
void    EqnFree(Eqn_p junk);

#define EqnSetProp(eqn, prop)  SetProp((eqn), (prop))
#define EqnDelProp(eqn, prop)  DelProp((eqn), (prop))
#define EqnFlipProp(eqn, prop) FlipProp((eqn), (prop))

/* Are _all_ properties in prop set in eqn? */
#define EqnQueryProp(eqn, prop) QueryProp((eqn), (prop))

/* Are any properties in prop set in eqn? */
#define EqnIsAnyPropSet(eqn, prop) IsAnyPropSet((eqn), (prop))

#define EqnIsOriented(eq) EqnQueryProp((eq), EPIsOriented)
#define EqnIsPositive(eq) EqnQueryProp((eq), EPIsPositive)
#define EqnIsEquLit(eq)   EqnQueryProp((eq), EPIsEquLiteral)
#define EqnIsMaximal(eq)  EqnQueryProp((eq), EPIsMaximal)
#define EqnIsStrictlyMaximal(eq)\\
                          EqnQueryProp((eq), EPIsStrictlyMax)

#define EqnHasEquiv(eq)  EqnQueryProp((eq), EPHasEquiv)
#define EqnIsDominated(eq)  EqnQueryProp((eq), EPIsDominated)
#define EqnDominates(eq)  EqnQueryProp((eq), EPDominates)

#define EqnIsGround(eq) \
        (TermIsGround((eq)->lterm) && TermIsGround((eq)->rterm))
Eqn_p   EqnParse(Scanner_p in, TB_p bank);
void    EqnPrint(FILE* out, Eqn_p eq, bool negated, bool fullterms);
#define EqnPrintOriginal(out, eq) \
        EqnPrint((out), (eq), normal, true)

void    EqnSwapSidesSimple(Eqn_p eq);
void    EqnSwapSides(Eqn_p eq);
void    EqnRecordTermDates(Eqn_p eq);

Eqn_p   EqnCopy(Eqn_p eq, TB_p bank);

#define EqnIsTrivial(eq, EqualTest) \
        EqualTest((eq)->lterm, (eq)->rterm)

bool    EqnEqualDirected(Eqn_p eq1,  Eqn_p eq2, TermEqualTestFun EqualTest);
bool    EqnEqual(Eqn_p eq1,  Eqn_p eq2, TermEqualTestFun EqualTest);
#define LiteralEqual(eq1, eq2, EqualTest) \
        (PropsAreEquiv((eq1),(eq2),EPIsPositive) &&\
	 EqnEqual((eq1),(eq2),(EqualTest)))

bool    EqnSubsumeDirected(Eqn_p subsumer, Eqn_p subsumed, Subst_p
			   subst, TermEqualTestFun EqualTest);
bool    EqnSubsume(Eqn_p subsumer, Eqn_p subsumed, Subst_p subst,
		   TermEqualTestFun EqualTest);
bool    EqnSubsumeP(Eqn_p subsumer, Eqn_p subsumed, 
		    TermEqualTestFun EqualTest);

bool    LiteralSubsumeP(Eqn_p subsumer, Eqn_p subsumed, 
			TermEqualTestFun EqualTest);

#define EqnEquiv(eq1, eq2, EqualTest) \
           (EqnSubsumeP((eq1),(eq2),(EqualTest)) \
	 &&(EqnSubsumeP((eq2),(eq1),(EqualTest))))

#define LiteralEquiv(eq1, eq2, EqualTest) \
        (((eq1)->positive == (eq2)->positive) \
	 && EqnEquiv((eq1),(eq2),(EqualTest)))

bool    EqnUnifyDirected(Eqn_p eq1, Eqn_p eq2, Subst_p subst);
bool    EqnUnify(Eqn_p eq1, Eqn_p eq2, Subst_p subst);
bool    EqnUnifyP(Eqn_p eq1, Eqn_p eq2);

bool          EqnOrient(OCB_p ocb, Eqn_p eq);

CompareResult EqnCompare(OCB_p ocb, Eqn_p eq1, Eqn_p eq2);
bool          EqnGreater(OCB_p ocb, Eqn_p eq1, Eqn_p eq2);
CompareResult LiteralCompare(OCB_p ocb, Eqn_p eq1, Eqn_p eq2);
bool          LiteralGreater(OCB_p ocb, Eqn_p eq1, Eqn_p eq2);

FunCode SubstNormEqn(Eqn_p eq, Subst_p subst, VarBank_p vars);

double  EqnWeight(Eqn_p eq, double max_multiplier, long vweight, long
		  fweight);
double  LiteralWeight(Eqn_p eq, double max_term_multiplier, double
		      max_literal_multiplier, double
		      pos_multiplier, long vweight, long fweight, bool
		      count_eq_encoding);

#define EqnCountMaximalLiterals(eqn) (EqnIsOriented(eqn)?1:2)

#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
