/*-----------------------------------------------------------------------

File  : che_heuristics.c

Author: Stephan Schulz

Contents
 
  High-Level interface functions to the heuristics module.


  Copyright 1998, 1999 by the author.
  This code is released under the GNU General Public Licence.
  See the file COPYING in the main CLIB directory for details.
  Run "eprover -h" for contact information.

Changes

<1> Mon Jun  8 02:14:51 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_heuristics.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


HeuristicAssocCell HeuristicsTable[]=
{
   {HEU_AUTO_MODE,    "Auto",       HCBAutoModeCreate},
   {HEU_NO_HEURISTIC, NULL,       NULL}
};

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: HCBAutoModeCreate()
//
//   Analyse the proof problem and return an hopefully suitable
//   heuristic. 
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

HCB_p HCBAutoModeCreate(HCBARGUMENTS)
{
   char *res = "Default";
   SpecFeature_p spec = &(control->problem_specs);

   OUTPRINT(1, "\n# Auto-Mode is analysing problem.\n");
   
   if(SpecAxiomsAreUnit(spec))
   {
      OUTPRINT(1, "# Axioms are unit-clauses.\n");
      if(SpecGoalsAreGround(spec))
      {
	 OUTPRINT(1, "# All goals are ground.\n");
	 SpecFeaturesAddEval(spec, 3,6,true);
	 if(SpecManyNGPosUnits(spec))
	 {
	    OUTPRINT(1, "# Many non-unit-axioms.\n");
	    res = "huY";
	 }
	 else if(SpecSomeNGPosUnits(spec))
	 {
	    OUTPRINT(1, "# Some non-unit-axioms.\n");
	    res = "huH";
	 }
	 else 
	 {
	    OUTPRINT(1, "# Few non-unit-axioms.\n");
	    if(SpecGoalsAreUnit(spec))
	    {
	       res = "huZ";
	    }
	    else
	    {
	       res = "hu5";
	    }
	 }	 
      }
      else
      {
	 OUTPRINT(1, "# Goals contain variables.\n");
	 res = "hnQ";
      }
   }
   else if(SpecAxiomsAreHorn(spec))
   {
      OUTPRINT(1, "# Axioms are horn-clauses.\n");
      SpecFeaturesAddEval(spec,
			  FEW_ABSDEFAULT,MANY_ABSDEFAULT,true);
      if(SpecFewNGPosUnits(spec)&&
	 SpecGoalsAreGround(spec)&&
	 SpecGoalsAreUnit(spec))
      {
	 OUTPRINT(1, "# Few pos-ground-units, ground goals.\n");
	 if(!control->problem_specs.eq_clauses)
	 {	    
	    OUTPRINT(1, "# No equality.\n");
	    res = "hh8";
	 }
	 else
	 {
	    OUTPRINT(1, "# There is equality.\n");
	    res ="hh32";
	 }
      }
      else if(SpecSomeNGPosUnits(spec)&&
	      SpecGoalsAreGround(spec)&&
	      SpecSomeEq(spec))
      {
	 OUTPRINT(1,
		  "# Some pos-ground-units, " 
		  "ground goals, some equality.\n");
	 res = "hhL0";
	 control->paramod_with_units_only = true;
      }
      else if(SpecManyNGPosUnits(spec)&&
	      SpecGoalsAreGround(spec)&&
	      (SpecNoEq(spec)||SpecPureEq(spec)))
      {
	 OUTPRINT(1,
		  "# Many pos-ground-units, " 
		  "ground goals, pure or no equality.\n");
	 res = "hh9";
	 control->paramod_with_units_only = true;
      }
      else if(!SpecGoalsAreGround(spec) &&
	      SpecGoalsAreUnit(spec) &&	      
	      ((SpecFewNGPosUnits(spec)&&
		SpecNoEq(spec))
	       ||
	       (SpecSomeNGPosUnits(spec)&&
		SpecSomeEq(spec))))	       
      {
	 res = "hh2";
      }
      else if((SpecGoalsAreHorn(spec) &&
	       SpecNoEq(spec) &&
	       SpecFewNGPosUnits(spec) &&
	       !SpecGoalsAreGround(spec))
	      ||
	      (SpecGoalsAreHorn(spec) &&
	       SpecSomeEq(spec)&&
	       SpecManyNGPosUnits(spec) &&
	       !SpecGoalsAreGround(spec))
	      ||
	      (SpecGoalsAreUnit(spec) &&
	       SpecNoEq(spec) &&
	       SpecSomeNGPosUnits(spec) &&
	       SpecGoalsAreGround(spec)))
      {
	 OUTPRINT(1, "# Really weird class....\n");
	 res = "hh23";
	 control->paramod_with_units_only = true;  
      }
      else
      {
	 OUTPRINT(1, "# Horn-default.\n");
	 res="hh21";
	 control->paramod_with_units_only = true;  	 
      }
   }
   else
   {
      OUTPRINT(1, "# Axioms are general clauses.\n");      
      if(SpecNoEq(spec))
      {
	 OUTPRINT(1, "# No equality.\n");      
	 res = "hgP";
      }
      else
      {
	 if(!SpecPureEq(spec) &&
	    !SpecGoalsAreGround(spec))
	 {
	    OUTPRINT(1, "# Some equality, goal has variables.\n");
	    
	    res = "hgQ";
	 }
	 else
	 {
	    res = "hgE";
	 }
      }
   }

   if(OutputLevel)
   {
      fprintf(stdout, "# Auto-Mode selected heuristic %s.\n", res);
   }

   return GetHeuristic(res, state, control, parms);
}


/*-----------------------------------------------------------------------
//
// Function: HCBCreate() 
//
//   Given a name and the necessary parameters, return a HCB. HCB's
//   must be registered in control->hcbs!
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

HCB_p HCBCreate(char* name, HCBARGUMENTS)
{
   int    i;
   DStr_p err;

   assert(control->ocb);

   for(i=0; HeuristicsTable[i].heuristic; i++)
   {
      if(strcmp(HeuristicsTable[i].name, name)==0)
      {
	 return HeuristicsTable[i].heuristic_create(state,control,parms);
      }
   }
   err = DStrAlloc();
   DStrAppendStr(err, "Heuristic \"");
   DStrAppendStr(err, name);
   DStrAppendStr(err, "\" unknown");
   Error(DStrView(err), USAGE_ERROR);
   DStrFree(err);
   
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: GetHeuristic()
//
//   Given a string (either a name or a Heuristic-Definition), return
//   a corresponding HCB.
//
// Global Variables: -
//
// Side Effects    : May create HCB's and WFCB's
//
/----------------------------------------------------------------------*/

HCB_p GetHeuristic(char* source, HCBARGUMENTS)
{
   char*     name;
   Scanner_p in;
   HCB_p     res;

   /* StreamTypeOptionString is a hack here...it will give wrong error
      messages for internal strings (but the function will not be
      called with internal strings anyways...) */
   in=CreateScanner(StreamTypeOptionString, source, true, NULL);
   if(TestInpTok(in, OpenBracket))
   {
      HeuristicDefParse(control->hcbs, in, control->wfcbs,
			control->ocb, state);
      name = SecureStrdup("Default");
   }
   else
   {      
      name = SecureStrdup(DStrView(AktToken(in)->literal));
      AcceptInpTok(in, Identifier);
   }
   DestroyScanner(in);
   res = HCBAdminFindHCB(control->hcbs, name);
   if(!res)
   {
      res = HCBCreate(name, state, control, parms);
      assert(res);
   }
   FREE(name);
   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


